#!/bin/bash

. /var/packages/MailPlus-Server/target/scripts/daemon/util.sh

ANTI_VIRUS_ENGINE="bitdefender"
BDAMSERVER_PID="/var/run/mailplus_server/bdamserver.pid"
BDAMSERVER="${MAIL_SERVER_PKG}/target/bin/bdamserver"
BDAMSERVER_CONF="${MAIL_SERVER_PKG}/target/etc/bdamserver.conf"
BDAMSERVER_LIB_DIR="${MAIL_SERVER_PKG}/target/lib/bitdefender"
BDAMSERVER_TMP_DIR="/tmp/mailplus_server/bitdefender"
SYNO_BITDEFENDER_TOOL="${MAIL_SERVER_PKG}/target/bin/syno_bitdefender_tool"

ANTIVIRUS_STATUS="/tmp/MailPlusServer.antivirus.status"

checkStarted()
{
	local pid process
	if [[ -f ${BDAMSERVER_PID} ]]; then
		pid="$(cat ${BDAMSERVER_PID})"
		# Add -a to let grep search through binary content
		process="$(grep -a bdamserver "/proc/${pid}/cmdline")"
	fi

	if [[ -n "${process}" ]]; then
		return 1
	fi
	return 0
}

bin_status()
{
	checkPidFileStatus "bdamserver" "${BDAMSERVER_PID}"
}

conf_status()
{
	local anti_virus_enable anti_virus_engine
	checkConfKey "anti_virus_enable"
	anti_virus_enable=$?
	anti_virus_engine="$(${MAILPLUS_SERVER_BACKEND_BINARY} --getConfKeyVal "anti_virus_engine")"

	if [[ "${RUNKEY_ENABLE}" == "${anti_virus_enable}" ]] && [[ "${ANTI_VIRUS_ENGINE}" == "${anti_virus_engine}" ]]; then
		return "${RUNKEY_ENABLE}"
	else
		return "${RUNKEY_DISABLE}"
	fi
}

status()
{
	conf_status
	local bdamserver_enable=$?
	bin_status
	local bdamserver_status=$?

	# TODO: Enable status check
	if [[ "${SERVICE_NOT_RUNNING}" -eq ${bdamserver_status} ]] && [[ "${RUNKEY_DISABLE}" -ne "${bdamserver_enable}" ]]; then
		return "${SERVICE_UNKNOWN}"
	elif [[ "${SERVICE_RUNNING}" -eq ${bdamserver_status} ]] && [[ "${RUNKEY_ENABLE}" -ne "${bdamserver_enable}" ]]; then
		return "${SERVICE_UNKNOWN}"
	fi
	return ${bdamserver_status}
}

clear_status_cache()
{
	rm -f "${ANTIVIRUS_STATUS}"
}

checkStarting()
{
	# When starting, there's bdamserver process, but no pid file
	[[ -n "$(getpid)" ]] && [[ ! -f "${BDAMSERVER_PID}" ]]
}

runDaemonInBackground()
{
	if checkStarting; then
		# There's another bdamserver starting
		debug_log "another bdamserver starting, quit"
		return
	fi

	env "LD_LIBRARY_PATH=${BDAMSERVER_LIB_DIR}" "${BDAMSERVER}" -c "${BDAMSERVER_CONF}"
	status
	daemon_status=$?
	debug_log "background daemon_status bdamserver: ${daemon_status}"
	if [[ ${daemon_status} -eq "${SERVICE_RUNNING}" ]]; then
		warn_log "BitDefender is started"
		generateInfoCache
	else
		err_log "BitDefender failed to start"
	fi
	clear_status_cache
}

runDaemon()
{
	mkdir -p "${BDAMSERVER_TMP_DIR}"
	runDaemonInBackground &

	local daemon_status=""
	for _ in $(seq 0 1 "${MAX_TRY_TIMES}")
	do
		status
		daemon_status=$?
		debug_log "daemon_status bdamserver: ${daemon_status}"
		if [[ ${daemon_status} -eq "${SERVICE_RUNNING}" ]]; then
			# binary is ready
			debug_log "foreground BitDefender is started"
			return "${START_SUCCESSFUL}"
		elif [[ ${daemon_status} -eq "${SERVICE_DEAD_FPID}" ]]; then
			# [FIXME] Maybe we can do something to recover
			return "${START_FAILED}"
		else
			sleep "${SLEEP_TIME}"
		fi
	done
	return "${START_FAILED}"
}

start()
{
	checkStarted
	if [ 1 -eq $? ] ; then
		echo "bdamserver is already running"
		clear_status_cache
	else
		runDaemon
	fi
}

checkStopped()
{
	for _ in $(seq 0 1 "${MAX_TRY_TIMES}")
	do
		if ! pidof bdamserver > /dev/null; then
			# bdamserver is terminated
			break
		fi
		sleep "${SLEEP_TIME}"
	done

	# if bdamserver still exists, force stop it
	if pidof bdamserver > /dev/null; then
		killall -9 bdamserver
	fi
}

stopDaemon()
{
	killByPidFile "TERM" "${BDAMSERVER_PID}"
	checkStopped
	warn_log "BitDefender is stopped"
}

stop()
{
	stopDaemon
	rm ${BDAMSERVER_PID} &> /dev/null
}

restart()
{
	stop
	sleep 1
	start
}

getpid()
{
	# bdamserver preforked child has the same cmdline with parent.
	# so this will return all the child process pids along with parent process pid.
	ps -C 'bdamserver' -o pid= -o args= | grep 'bdamserver' | awk '{print $1}'
}

reload()
{
	if [[ -f ${BDAMSERVER_PID} ]]; then
		killByPidFile "HUP" "${BDAMSERVER_PID}"
	else
		local pids
		pids="$(getpid)"

		if [ -n "${pids}" ]; then
			for pid in ${pids}; do
				killByPid "HUP" "${pid}"
			done
		else
			err_log "bdamserver is not alive"
		fi
	fi
}

# For manual update only
update()
{
	if isLoadBalancer; then
		nowTime=$(date "+%Y/%m/%d %H:%M:%S")
		${MAILPLUS_SERVER_BACKEND_BINARY} --setConfKeyVal "anti_virus_last_update_time" "${nowTime}"
	fi
}

# For manual update only
update_local_only()
{
	"${SYNO_BITDEFENDER_TOOL}" start_update

	# Clear antivirus updating status regardless of finishing update, because BitDefender can scan while updating
	clear_status_cache
}

composePID()
{
	if [[ -f "${BDAMSERVER_PID}" ]]; then
		local pid
		pid=$(sed -e 's/^[[:space:]]*//' -e 's/[[:space:]]*$//' "${BDAMSERVER_PID}")
		echo "${pid}"
	fi
}

composeScript()
{
	echo "/var/packages/MailPlus-Server/target/scripts/daemon/bdamserver.sh"
}

generateInfoCache()
{
	"${SYNO_BITDEFENDER_TOOL}" generate_info_cache
}

case "$1" in
	start)
		startOperation "mailserver_service_bdamserver"
		;;
	restart)
		restartOperation "mailserver_service_bdamserver"
		;;
	stop)
		stopOperation "mailserver_service_bdamserver"
		;;
	reload)
		reloadOperation "mailserver_service_bdamserver"
		;;
	status)
		status
		;;
	update)
		update
		;;
	update_local_only)
		update_local_only
		;;
	*)
		echo "Usage: $0 {start|stop|restart|status|update|reload|update_local_only}"
esac
