#!/bin/bash

. /var/packages/MailPlus-Server/target/scripts/daemon/util.sh

RSPAMD_PID="/var/run/mailplus_server/rspamd.pid"
RSPAMD="${MAIL_SERVER_PKG}/target/bin/rspamd"

RSPAMD_STATUS_CACHE="/tmp/MailPlusServer.rspamd.status"

RSPAMD_MAX_TRY_TIMES=30

findProcess()
{
	ps -C rspamd -o pid,args | grep 'rspamd: main'
}

findProcessByPid()
{
	local pid="$1"
	ps -p "${pid}" -o pid,args | grep 'rspamd: main'
}

getpid()
{
	findProcess | awk '{print $1}'
}

checkStarted()
{
	if [ -f ${RSPAMD_PID} ]; then
		local pid="$(cat ${RSPAMD_PID})"

		if [ -n "$(findProcessByPid "${pid}")" ]; then
			return 1
		fi
	fi

	return 0
}

bin_status()
{
	checkPidFileStatus "rspamd" "${RSPAMD_PID}"
}

conf_status()
{
	checkConfKey "anti_virus_enable"
	if [ "${RUNKEY_ENABLE}" -eq $? ]; then
		return "${RUNKEY_ENABLE}"
	fi
	checkConfKey "spam_enable"
	if [ "${RUNKEY_ENABLE}" -eq $? ]; then
		return "${RUNKEY_ENABLE}"
	fi
	checkConfKey "mcp_enable"
	if [ "${RUNKEY_ENABLE}" -eq $? ]; then
		return "${RUNKEY_ENABLE}"
	fi
	return "${RUNKEY_DISABLE}"
}

status()
{
	conf_status
	local rspamd_enable=$?
	bin_status
	local rspamd_status=$?

	if [ "${SERVICE_NOT_RUNNING}" -eq ${rspamd_status} ] && [ "${RUNKEY_DISABLE}" -ne "${rspamd_enable}" ]; then
		return "${SERVICE_UNKNOWN}"
	elif [ "${SERVICE_RUNNING}" -eq ${rspamd_status} ] && [ "${RUNKEY_ENABLE}" -ne "${rspamd_enable}" ]; then
		return "${SERVICE_UNKNOWN}"
	fi
	return ${rspamd_status}
}

clear_status_cache()
{
	rm -f "${RSPAMD_STATUS_CACHE}"
}

runDaemon()
{
	local exit_status
	${RSPAMD} -u MailPlus-Server -g system
	exit_status=$?

	if [[ "${exit_status}" -ne 0 ]]; then
		err_log "Failed to execute Rspamd, exit status: ${exit_status}"
	fi

	local daemon_status=""
	for i in $(seq 0 1 "${RSPAMD_MAX_TRY_TIMES}")
	do
		status
		daemon_status=$?
		debug_log "daemon_status rspamd: ${daemon_status}"
		if [ ${daemon_status} -eq "${SERVICE_RUNNING}" ]; then
			# binary is ready
			warn_log "Rspamd is started"
			clear_status_cache
			return ${START_SUCCESSFUL}
		elif [ ${daemon_status} -eq "${SERVICE_DEAD_FPID}" ]; then
			# [FIXME] Maybe we can do something to recover
			return ${START_FAILED}
		else
			sleep "${SLEEP_TIME}"
		fi
	done
	return ${START_FAILED}
}

start()
{
	checkStarted
	if [ 1 -eq $? ] ; then
		echo "rspamd is already running"
		clear_status_cache
	else
		runDaemon
	fi
}

checkStopped()
{
	for i in $(seq 0 1 "${RSPAMD_MAX_TRY_TIMES}")
	do
		if ! pidof rspamd > /dev/null; then
			# rspamd is terminated
			break
		fi
		sleep "${SLEEP_TIME}"
	done

	# if rspamd still exists, force stop it
	if pidof rspamd > /dev/null; then
		killall -9 rspamd
	fi
}

stopDaemon()
{
	killByPidFile "TERM" "${RSPAMD_PID}"
	checkStopped
	warn_log "Rspamd is stopped"
}

stop()
{
	stopDaemon
	rm ${RSPAMD_PID} &> /dev/null
}

restart()
{
	stop
	sleep 1
	start
}

reload()
{
	if [ -f ${RSPAMD_PID} ]; then
		killByPidFile "HUP" "${RSPAMD_PID}"
	else
		local pid="$(getpid)"

		if [ -n "${pid}" ]; then
			killByPid "HUP" "${pid}"
		else
			err_log "rspamd is not alive"
		fi
	fi
}

composePID()
{
	local pid=$(cat ${RSPAMD_PID} | sed -e 's/^[[:space:]]*//' -e 's/[[:space:]]*$//')
	echo "${pid}"
}

composeScript()
{
	echo "/var/packages/MailPlus-Server/target/scripts/daemon/rspamd.sh"
}

case "$1" in
	start)
		startOperation "mailserver_service_rspamd"
		;;
	restart)
		restartOperation "mailserver_service_rspamd"
		;;
	stop)
		stopOperation "mailserver_service_rspamd"
		;;
	reload)
		reloadOperation "mailserver_service_rspamd"
		;;
	status)
		status
		;;
esac
