--[[
Copyright (c) 2000-2020 Synology Inc. All rights reserved.
]]--

--[[
-- @module bitdefender
-- This module contains bitdefender access functions
]]--

local lua_util = require "lua_util"
local tcp = require "rspamd_tcp"
local upstream_list = require "rspamd_upstream_list"
local rspamd_util = require "rspamd_util"
local rspamd_logger = require "rspamd_logger"
local common = require "lua_scanners/common"
local fun = require "fun"

local N = "bitdefender"

local default_message = '${SCANNER}: virus found: "${VIRUS}"'

-- Non-exhaustive
local BITDEFENDER_SCAN_OPTIONS = {
  ARCHIVES  = 0x01,
  PACKED    = 0x02,
  EMAILS    = 0x04,
  SPAMCHECK = 0x40,
}

-- luacheck: globals bit
-- `bit` is bitwise operation library in Lua 5.1
-- TODO: Consider not using spam option if spam is disabled
local BITDEFENDER_DEFAULT_SCAN_OPTION = bit.bor(
  BITDEFENDER_SCAN_OPTIONS.ARCHIVES,
  BITDEFENDER_SCAN_OPTIONS.PACKED,
  BITDEFENDER_SCAN_OPTIONS.EMAILS,
  BITDEFENDER_SCAN_OPTIONS.SPAMCHECK
)

local SYNOLOGY_THREAT_TYPE_SPAM  = 'spam'
local SYNOLOGY_THREAT_TYPE_VIRUS = 'virus'
local SYNOLOGY_THREAT_TYPE_SKIP  = 'skip'

local SYNOLOGY_SPAM_ACTION_SPAM = 'spam'
local SYNOLOGY_SPAM_ACTION_INBOX = 'inbox'

local BITDEFENDER_THREAT_TYPES = {
  V = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Virus' },
  S = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Spyware' },
  W = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Adware' },
  D = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Dialer' },
  A = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Potentially Dangerous App' },
  M = { type = SYNOLOGY_THREAT_TYPE_SPAM,  name = 'Spam' },
  P = { type = SYNOLOGY_THREAT_TYPE_SPAM,  name = 'Phishing' },
  B = { type = SYNOLOGY_THREAT_TYPE_VIRUS, name = 'Need Sandbox' },
  E = { type = SYNOLOGY_THREAT_TYPE_SPAM,  name = 'Marketing Email' },
  U = { type = SYNOLOGY_THREAT_TYPE_SPAM,  name = 'Bulk Spam' },
}

local BITDEFENDER_SPAM_TYPE_ACTION_NAMES = {
  M = 'spam',
  P = 'phishing',
  E = 'marketing_email',
  U = 'bulk_spam',
}

-- mapping from return code to boolean determine if extra parameters exist
local BITDEFENDER_SCAN_RETURN_CODES = {
  [227] = { clean = true,  error = false, has_threat_type = false, has_malware_name = false, message = 'CLEAN' },
  [226] = { clean = false, error = false, has_threat_type = true,  has_malware_name = true,  message = 'DISINFECTED' },
  [220] = { clean = false, error = false, has_threat_type = true,  has_malware_name = true,  message = 'DISINFECTFAILED' },
  [222] = { clean = false, error = false, has_threat_type = true,  has_malware_name = true,  message = 'INFECTED' },
  [223] = { clean = false, error = false, has_threat_type = false, has_malware_name = true,  message = 'SUSPICIOUS' },
  [225] = { clean = false, error = true,  has_threat_type = false, has_malware_name = false, message = 'ENCRYPTED' },
  [221] = { clean = false, error = true,  has_threat_type = false, has_malware_name = false, message = 'FAILED' },
  [224] = { clean = false, error = true,  has_threat_type = false, has_malware_name = false, message = 'CORRUPTED' }
}

-- Command
local BITDEFENDER_SCANFILE_COMMAND = 'SCANFILE'

-- Response format
local RETURN_CODE_LENGTH = 3
local EXTRA_PARAMETER_OFFSET = 5
local THREAT_TYPE_LENGTH = 1

local function bitdefender_config(opts)
  local bitdefender_conf = {
    name = N,
    scan_mime_parts = false,
    scan_text_mime = false,
    scan_image_mime = false,
    default_port = 1344,
    log_clean = false,
    timeout = 5.0, -- FIXME: this will break task_timeout!
    detection_category = {
      spam = "spam",
      virus = "virus",
      skip = "skip",
    },
    retransmits = 2,
    message = default_message,
    tmpdir = '/tmp',
    spam_action = {
      spam = SYNOLOGY_SPAM_ACTION_SPAM,
      phishing = SYNOLOGY_SPAM_ACTION_SPAM,
      marketing_email = SYNOLOGY_SPAM_ACTION_INBOX, -- default skip marketing email
      bulk_spam = SYNOLOGY_SPAM_ACTION_SPAM,
    },
  }

  -- Load bitdefender antivirus config
  bitdefender_conf = lua_util.override_defaults(bitdefender_conf, opts)
  -- Load bitdefender spam config
  bitdefender_conf = lua_util.override_defaults(bitdefender_conf, rspamd_config:get_all_opt('bitdefender'))

  if not bitdefender_conf.prefix then
    bitdefender_conf.prefix = 'rs_' .. bitdefender_conf.name .. '_'
  end

  if not bitdefender_conf.log_prefix then
    if bitdefender_conf.name:lower() == bitdefender_conf.type:lower() then
      bitdefender_conf.log_prefix = bitdefender_conf.name
    else
      bitdefender_conf.log_prefix = bitdefender_conf.name .. ' (' .. bitdefender_conf.type .. ')'
    end
  end

  if not bitdefender_conf['servers'] then
    rspamd_logger.errx(rspamd_config, 'no servers defined')

    return nil
  end

  bitdefender_conf['upstreams'] = upstream_list.create(rspamd_config,
    bitdefender_conf['servers'],
    bitdefender_conf.default_port)

  if bitdefender_conf['upstreams'] then
    lua_util.add_debug_alias('antivirus', bitdefender_conf.name)
    return bitdefender_conf
  end

  rspamd_logger.errx(rspamd_config, 'cannot parse servers %s',
    bitdefender_conf['servers'])
  return nil
end

local function generate_temp_filename(tmpdir)
  if not rspamd_util.mkdir(tmpdir, true) then
    rspamd_logger.errx('cannot create temp directory: %s', tmpdir)
    return nil
  end

  local filename
  repeat
    filename = string.format('%s/%s.tmp', tmpdir, rspamd_util.random_hex(32))
  until not rspamd_util.file_exists(filename)

  return filename
end

local function bitdefender_request_new(command, option, path)
  return string.format('%s %s %s\r\n', command, option, path)
end

-- strip prefix from every line and joined as a string
local function bitdefender_strip_prefix(response, prefix_length)
  local responses = lua_util.str_split(response, '\r\n')

  return fun.reduce(function(acc, response_line)
    return acc .. response_line
  end, '', fun.map(function(response_line)
    return response_line:sub(prefix_length + 1)
  end, responses))
end

-- BitDefender response:
--
-- 1. Return code indicates error:
--
--           3 bytes            1 byte       remaining bytes
-- +--------+--------+--------+--------+--------+--------+--------+
-- |                          | conti- |                          |
-- |       return code        | nuity  |          error           |
-- |                          | symbol |                          |
-- +--------+--------+--------+--------+--------+--------+--------+
--
-- 2. Return code indicates there is malware name
--
--           3 bytes            1 byte       remaining bytes
-- +--------+--------+--------+--------+--------+--------+--------+
-- |                          | conti- |                          |
-- |       return code        | nuity  |       malware name       |
-- |                          | symbol |                          |
-- +--------+--------+--------+--------+--------+--------+--------+
--
-- 3. Return code indicates there are threat type and malware name
--
--           3 bytes            1 byte   1 byte   1 byte       remaining bytes
-- +--------+--------+--------+--------+--------+--------+--------+--------+--------+
-- |                          | conti- |        |        |                          |
-- |       return code        | nuity  | threat | space  |       malware name       |
-- |                          | symbol |  type  |        |                          |
-- +--------+--------+--------+--------+--------+--------+--------+--------+--------+
--
-- Skip continuity symbol as we don't enable progress information
--
-- return: error, return code, threat type, malware name/error message
local function bitdefender_response_parse(response)
  local return_code, return_code_attributes, threat_type, malware_name

  return_code = tonumber(response:sub(1, RETURN_CODE_LENGTH))

  -- Handle 3XX, 4XX, 5XX return code
  if (return_code >= 300 and return_code <= 399) or
    (return_code >= 400 and return_code <= 499) or
    (return_code >= 500 and return_code <= 599) then
    local error_message = bitdefender_strip_prefix(response, EXTRA_PARAMETER_OFFSET - 1)
    return true, return_code, nil, error_message
  end

  return_code_attributes = BITDEFENDER_SCAN_RETURN_CODES[return_code]
  if return_code_attributes.error then
    return true, return_code, nil, return_code_attributes.message
  end

  if return_code_attributes.has_threat_type then
    threat_type = response:sub(EXTRA_PARAMETER_OFFSET, EXTRA_PARAMETER_OFFSET + THREAT_TYPE_LENGTH - 1)
  end

  if return_code_attributes.has_malware_name then
    local malware_name_offset = EXTRA_PARAMETER_OFFSET
    if return_code_attributes.has_threat_type then
      malware_name_offset = malware_name_offset + THREAT_TYPE_LENGTH + 1
    end

    malware_name = bitdefender_strip_prefix(response, malware_name_offset - 1)
  end

  return nil, return_code, threat_type, malware_name
end

local function bitdefender_is_clean(return_code)
  return BITDEFENDER_SCAN_RETURN_CODES[return_code].clean
end

local function bitdefender_generate_scan_result(return_code, threat_type_vtype, threat_type_name, malware_name)
  local return_code_message = BITDEFENDER_SCAN_RETURN_CODES[return_code].message

  if bitdefender_is_clean(return_code) then
    return return_code_message
  end

  return string.format('%s - %s - %s - %s', return_code_message, threat_type_vtype, threat_type_name, malware_name)
end

local function bitdefender_get_action_type(threat_type)
  return BITDEFENDER_THREAT_TYPES[threat_type].type
end

local function bitdefender_get_action_name(threat_type)
  return BITDEFENDER_THREAT_TYPES[threat_type].name
end

local function bitdefender_get_spam_action(rule, threat_type)
  return rule.spam_action[BITDEFENDER_SPAM_TYPE_ACTION_NAMES[threat_type]]
end

local function bitdefender_check(task, content, digest, rule)
  -- Disable cache for BitDefender as it's anti-spam is cloud based and can get
  -- updated anytime.
  local upstream = rule.upstreams:get_upstream_round_robin()
  local addr = upstream:get_addr()
  local retransmits = rule.retransmits
  local is_threat = false
  local vmessage, vtype
  local fname = generate_temp_filename(rule.tmpdir)
  if not fname then
    rspamd_logger.errx('cannot create temp filename')
    return
  end

  local message_fd = rspamd_util.create_file(fname)
  local request

  if not message_fd then
    rspamd_logger.errx('cannot store file for bitdefender scan: %s', fname)
    return
  end

  if type(content) == 'string' then
    -- Create rspamd_text
    local rspamd_text = require "rspamd_text"
    content = rspamd_text.fromstring(content)
  end
  content:save_in_file(message_fd)

  -- Ensure file cleanup
  task:get_mempool():add_destructor(function()
    os.remove(fname)
    rspamd_util.close_file(message_fd)
  end)

  local function bitdefender_response_finish()
    if is_threat then
      common.yield_result(task, rule, vmessage, 1, false, vtype)
    else
      if rule['log_clean'] then
        rspamd_logger.infox(task, '%s [%s]: message is clean', rule['symbol'], rule['type'])
      else
        lua_util.debugm(rule.name, task, '%s: message or mime_part is clean', rule.log_prefix)
      end
    end
  end

  local function bitdefender_response_handler(data)
    rspamd_logger.infox(task, 'bitdefender_response_handler, response: "%s"', data)

    local error, return_code, threat_type, malware_name = bitdefender_response_parse(data)

    if error then
      local error_message = malware_name
      rspamd_logger.errx(task, 'failed to scan, return code: %d, error message: %s', return_code, error_message)
    else
      local vname

      -- Get vtype, vname
      if not bitdefender_is_clean(return_code) then
        vtype = bitdefender_get_action_type(threat_type)
        vname = bitdefender_get_action_name(threat_type)

        -- Check BitDefender spam action
        if vtype == SYNOLOGY_THREAT_TYPE_SPAM then
          if bitdefender_get_spam_action(rule, threat_type) == SYNOLOGY_SPAM_ACTION_INBOX then
            vtype = SYNOLOGY_THREAT_TYPE_SKIP
          end
        end
      end

      local message = bitdefender_generate_scan_result(return_code, vtype, vname, malware_name)
      rspamd_logger.infox(task, 'bitdefender_response_handler, message: %s', message)

      if not bitdefender_is_clean(return_code) then
        -- Check antivirus whitelist
        if vtype == SYNOLOGY_THREAT_TYPE_VIRUS and common.check_antivirus_whitelist(task) then
          is_threat = false
        else
          is_threat = true
          vmessage = message
        end
      end

      bitdefender_response_finish()
    end
  end

  local function bitdefender_callback(err, data, conn)
    if err then

      -- set current upstream to fail because an error occurred
      upstream:fail()

      -- retry with another upstream until retransmits exceeds
      if retransmits > 0 then

        retransmits = retransmits - 1

        -- Select a different upstream!
        upstream = rule.upstreams:get_upstream_round_robin()
        addr = upstream:get_addr()

        lua_util.debugm(rule.name, task, '%s: retry IP: %s',
          rule.log_prefix, addr)

        tcp.request({
            task = task,
            host = addr:to_string(),
            port = addr:get_port(),
            timeout = rule['timeout'],
            callback = bitdefender_callback,
            data = request,
          })
      else
        rspamd_logger.errx(task, '%s: failed to scan, maximum retransmits exceed', rule.log_prefix)
        common.yield_result(task, rule, 'failed to scan and retransmits exceed', 0.0, 'fail')
      end

    else
      upstream:ok()

      bitdefender_response_handler(tostring(data))
    end
  end

  if rule.dynamic_scan then
    local pre_check, pre_check_msg = common.check_metric_results(task, rule)
    if pre_check then
      rspamd_logger.infox(task, '%s: aborting: %s', rule.log_prefix, pre_check_msg)
      return true
    end
  end

  request = bitdefender_request_new(BITDEFENDER_SCANFILE_COMMAND, BITDEFENDER_DEFAULT_SCAN_OPTION, fname)

  tcp.request({
      task = task,
      host = addr:to_string(),
      port = addr:get_port(),
      timeout = rule['timeout'],
      callback = bitdefender_callback,
      data = request,
    })
end

return {
  type = 'antivirus',
  description = 'bitdefender antivirus',
  configure = bitdefender_config,
  check = bitdefender_check,
  name = N
}
