--[[
Copyright (c) 2000-2019 Synology Inc. All rights reserved.
]]--

--[[[
-- @module synoav_mcafee
-- This module contains synoav_mcafee access functions
--]]

local lua_util = require "lua_util"
local tcp = require "rspamd_tcp"
local upstream_list = require "rspamd_upstream_list"
local rspamd_util = require "rspamd_util"
local rspamd_logger = require "rspamd_logger"
local common = require "lua_scanners/common"
local ucl = require "ucl"

local N = "synoav_mcafee"

local default_message = '${SCANNER}: virus found: "${VIRUS}"'

local function synoav_mcafee_config(opts)
  local synoav_mcafee_conf = {
    name = N,
    scan_mime_parts = true,
    scan_text_mime = false,
    scan_image_mime = false,
    default_port = 3310,
    log_clean = false,
    timeout = 5.0, -- FIXME: this will break task_timeout!
    detection_category = "virus",
    retransmits = 2,
    cache_expire = 3600, -- expire redis in one hour
    message = default_message,
    tmpdir = '/tmp',
  }

  synoav_mcafee_conf = lua_util.override_defaults(synoav_mcafee_conf, opts)

  if not synoav_mcafee_conf.prefix then
    synoav_mcafee_conf.prefix = 'rs_' .. synoav_mcafee_conf.name .. '_'
  end

  if not synoav_mcafee_conf.log_prefix then
    if synoav_mcafee_conf.name:lower() == synoav_mcafee_conf.type:lower() then
      synoav_mcafee_conf.log_prefix = synoav_mcafee_conf.name
    else
      synoav_mcafee_conf.log_prefix = synoav_mcafee_conf.name .. ' (' .. synoav_mcafee_conf.type .. ')'
    end
  end

  if not synoav_mcafee_conf['servers'] then
    rspamd_logger.errx(rspamd_config, 'no servers defined')

    return nil
  end

  synoav_mcafee_conf['upstreams'] = upstream_list.create(rspamd_config,
      synoav_mcafee_conf['servers'],
      synoav_mcafee_conf.default_port)

  if synoav_mcafee_conf['upstreams'] then
    lua_util.add_debug_alias('antivirus', synoav_mcafee_conf.name)
    return synoav_mcafee_conf
  end

  rspamd_logger.errx(rspamd_config, 'cannot parse servers %s',
      synoav_mcafee_conf['servers'])
  return nil
end

local function generate_temp_filename(tmpdir)
  if not rspamd_util.mkdir(tmpdir, true) then
    rspamd_logger.errx('cannot create temp directory: %s', tmpdir)
    return nil
  end

  local filename
  repeat
    filename = string.format('%s/%s.tmp', tmpdir, rspamd_util.random_hex(32))
  until not rspamd_util.file_exists(filename)

  return filename
end

local function synoav_mcafee_check(task, content, digest, rule)
  local function synoav_mcafee_check_uncached ()
    local upstream = rule.upstreams:get_upstream_round_robin()
    local addr = upstream:get_addr()
    local retransmits = rule.retransmits
    local vnames = {}
    local fname = generate_temp_filename(rule.tmpdir)
    if not fname then
      rspamd_logger.errx('cannot create temp filename')
      return
    end

    local message_fd = rspamd_util.create_file(fname)
    local request

    if not message_fd then
      rspamd_logger.errx('cannot store file for synoav_mcafee scan: %s', fname)
      return
    end

    if type(content) == 'string' then
      -- Create rspamd_text
      local rspamd_text = require "rspamd_text"
      content = rspamd_text.fromstring(content)
    end
    content:save_in_file(message_fd)

    -- Ensure file cleanup
    task:get_mempool():add_destructor(function()
      os.remove(fname)
      rspamd_util.close_file(message_fd)
    end)


    local function synoav_mcafee_request_new(command, path, content)
      local body = ucl.to_format({
        command = command,
        path = path,
        content = content
      }, 'json')

      local header = rspamd_util.pack(">I4", string.len(body))

      return { header, body }
    end

    local function synoav_mcafee_response_finish()
      local cached
      if #vnames == 0 then
        cached = 'OK'
        if rule['log_clean'] then
          rspamd_logger.infox(task, '%s [%s]: message is clean', rule['symbol'], rule['type'])
        else
          lua_util.debugm(rule.name, task, '%s: message or mime_part is clean', rule.log_prefix)
        end
      else
        common.yield_result(task, rule, vnames)
        cached = vnames
      end

      if cached then
        common.save_av_cache(task, digest, rule, cached)
      end
    end

    local function synoav_mcafee_response_handler(data)
      rspamd_logger.infox(task, 'synoav_mcafee_response_handler, response: %s', data)

      -- parse_string will not reset parser state, which means subsequitential call
      -- to parse_string will only append the data to parser's chunk.
      -- Thus, use new parser for every parse_string call.
      local parser = ucl.parser()
      local res, err = parser:parse_string(data)

      if not res then
        rspamd_logger.errx(task, 'failed to parse response: %s, error: %s', data, err)
        common.yield_result(task, rule, 'unparsed response: ' .. data, 0.0, 'fail')
        return
      else
        local response = parser:get_object()

        if response['status'] == "failure" then
          rspamd_logger.errx(task, 'failed to scan, result: %s', response['result'])
          common.yield_result(task, rule, 'failed to scan: ' .. response['result'], 0.0, 'fail')
          return
        else
          local vname = string.match(response['result'], ' : ([^\n]*) FOUND')

          if vname then
            vnames[#vnames + 1] = vname
          end

          synoav_mcafee_response_finish()
        end
      end
    end

    local function synoav_mcafee_callback(err, data, conn)
      if err then

        -- set current upstream to fail because an error occurred
        upstream:fail()

        -- retry with another upstream until retransmits exceeds
        if retransmits > 0 then

          retransmits = retransmits - 1

          -- Select a different upstream!
          upstream = rule.upstreams:get_upstream_round_robin()
          addr = upstream:get_addr()

          lua_util.debugm(rule.name, task, '%s: retry IP: %s',
              rule.log_prefix, addr)

          tcp.request({
            task = task,
            host = addr:to_string(),
            port = addr:get_port(),
            timeout = rule['timeout'],
            callback = synoav_mcafee_callback,
            data = request,
          })
        else
          rspamd_logger.errx(task, '%s: failed to scan, maximum retransmits exceed', rule.log_prefix)
          common.yield_result(task, rule, 'failed to scan and retransmits exceed', 0.0, 'fail')
        end

      else
        upstream:ok()

        -- synoav_mcafee packet = { header, body }
        -- * header is 4 bytes network byte order integer contains body length
        -- * body is a JSON format string
        -- Since synoav_mcafee packet has no stop pattern, data may only contain header.
        -- In this case, we add another read event to get body.
        if data:len() == 4 then
          conn:add_read(function(err, data, conn)
            conn:close()
            synoav_mcafee_response_handler(tostring(data))
          end)
        else
          conn:close()
          synoav_mcafee_response_handler(string.sub(data:str(), 5))
        end
      end
    end

    if rule.dynamic_scan then
      local pre_check, pre_check_msg = common.check_metric_results(task, rule)
      if pre_check then
        rspamd_logger.infox(task, '%s: aborting: %s', rule.log_prefix, pre_check_msg)
        return true
      end
    end

    request = synoav_mcafee_request_new('file', fname, '')

    tcp.request({
      task = task,
      host = addr:to_string(),
      port = addr:get_port(),
      timeout = rule['timeout'],
      callback = synoav_mcafee_callback,
      data = request,
    })
  end

  if common.need_av_check(task, content, rule) then
    if common.check_av_cache(task, digest, rule, synoav_mcafee_check_uncached) then
      return
    else
      synoav_mcafee_check_uncached()
    end
  end
end

return {
  type = 'antivirus',
  description = 'synoav_mcafee antivirus',
  configure = synoav_mcafee_config,
  check = synoav_mcafee_check,
  name = N
}
