--[[
Copyright (c) 2000-2020 Synology Inc. All rights reserved.
]]--

local exports = {}
local constants = require "synology/constants"

exports.is_prefix = function(str, prefix)
  return string.sub(str, 1, string.len(prefix)) == prefix
end

exports.is_mcp_symbol = function(symbol)
  return exports.is_prefix(symbol['name'], constants.MCP_SYMBOL_PREFIX)
end

exports.antivirus_symbols = function()
  local symbols = {}

  for _, antivirus_symbol in ipairs(constants.ANTIVIRUS_SYMBOLS) do
    for _, antivirus_symbol_suffix in ipairs(constants.ANTIVIRUS_SYMBOL_SUFFIXES) do
      symbols[#symbols + 1] = antivirus_symbol .. antivirus_symbol_suffix
    end
  end

  return symbols
end

exports.is_antivirus_symbol = function(symbol)
  for _, antivirus_symbol in ipairs(exports.antivirus_symbols()) do
    if symbol['name'] == antivirus_symbol then
      return true
    end
  end

  return false
end

exports.antivirus_fail_symbols = function()
  local symbols = {}

  for _, antivirus_symbol in ipairs(constants.ANTIVIRUS_SYMBOLS) do
    for _, antivirus_fail_symbol_suffix in ipairs(constants.ANTIVIRUS_FAIL_SYMBOL_SUFFIXES) do
      symbols[#symbols + 1] = antivirus_symbol .. antivirus_fail_symbol_suffix
    end
  end

  return symbols
end

exports.is_antivirus_fail_symbol = function(symbol)
  for _, antivirus_symbol in ipairs(exports.antivirus_fail_symbols()) do
    if symbol['name'] == antivirus_symbol then
      return true
    end
  end

  return false
end

exports.is_antivirus_whitelist_symbol = function(symbol)
  for _, antivirus_whitelist_symbol in ipairs(constants.ANTIVIRUS_WHITELIST_SYMBOLS) do
    if antivirus_whitelist_symbol == symbol['name'] then
      return true
    end
  end

  return false
end

exports.is_spam_symbol = function(symbol)
  return not exports.is_mcp_symbol(symbol) and
    not exports.is_antivirus_symbol(symbol) and
    not exports.is_antivirus_fail_symbol(symbol) and
    not exports.is_antivirus_whitelist_symbol(symbol)
end

exports.is_whitelist_symbol = function(symbol)
  local symbol_parts = rspamd_str_split(symbol['name'], '_')
  if #symbol_parts ~= 3 then
    return false
  end
  if symbol_parts[1] ~= constants.SPAM_WHITELIST_PREFIX then
    return false
  end
  local infix_match = false
  for _, spam_blackwhite_infix in ipairs(constants.SPAM_BLACKWHITE_INFIXES) do
    if spam_blackwhite_infix == symbol_parts[2] then
      infix_match = true
      break
    end
  end
  local suffix_match = false
  for _, spam_blackwhite_suffix in ipairs(constants.SPAM_BLACKWHITE_SUFFIXES) do
    if spam_blackwhite_suffix == symbol_parts[3] then
      suffix_match = true
      break
    end
  end
  return infix_match and suffix_match
end

exports.is_blacklist_symbol = function(symbol)
  local symbol_parts = rspamd_str_split(symbol['name'], '_')
  if #symbol_parts ~= 3 then
    return false
  end
  if symbol_parts[1] ~= constants.SPAM_BLACKLIST_PREFIX then
    return false
  end
  local infix_match = false
  for _, spam_blackwhite_infix in ipairs(constants.SPAM_BLACKWHITE_INFIXES) do
    if spam_blackwhite_infix == symbol_parts[2] then
      infix_match = true
      break
    end
  end
  local suffix_match = false
  for _, spam_blackwhite_suffix in ipairs(constants.SPAM_BLACKWHITE_SUFFIXES) do
    if spam_blackwhite_suffix == symbol_parts[3] then
      suffix_match = true
      break
    end
  end
  return infix_match and suffix_match
end

exports.is_dkim_spam_symbol = function(symbol)
  return symbol['name'] == constants.DKIM_SPAM_SYMBOL
end

exports.is_dmarc_quarantine_symbol = function(symbol)
  return symbol['name'] == constants.DMARC_QUARANTINE_SYMBOL
end

exports.is_bitdefender_spam_symbol = function(symbol)
  return symbol['name'] == constants.BITDEFENDER_SPAM_SYMBOL
end

exports.is_bitdefender_skip_symbol = function(symbol)
  return symbol['name'] == constants.BITDEFENDER_SKIP_SYMBOL
end

exports.is_bitdefender_clean_symbol = function(symbol)
  return symbol['name'] == constants.BITDEFENDER_CLEAN_SYMBOL
end

exports.generate_score_status = function(score, required_score)
  return string.format('score=%s, required %s', score, required_score)
end

exports.generate_full_score_status = function(score, required, message)
  return string.format('%s, %s', exports.generate_score_status(score, required), message)
end

exports.generate_ham_status = function(symbol, required_score)
  return exports.generate_full_score_status(0, required_score, symbol['name'] .. ' 0')
end

exports.generate_spam_status = function(symbol, required_score)
  return exports.generate_full_score_status(required_score, required_score, symbol['name'] .. ' ' .. required_score)
end

exports.generate_bitdefender_status = function(symbol)
  return symbol['name']
end

exports.generate_bitdefender_spam_status = function(symbol)
  return symbol['options'][1]
end

exports.generate_bitdefender_skip_status = function(symbol)
  return symbol['options'][1]
end

exports.generate_bitdefender_clean_status = function(symbol)
  return 'CLEAN'
end

return exports
