# Maildir repository support
# Copyright (C) 2002-2015 John Goerzen & contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

from offlineimap import folder
from offlineimap.ui import getglobalui
from offlineimap.error import OfflineImapError
from offlineimap.repository.Base import BaseRepository
import time
import os
from stat import *

import glob

import sys
for egg in glob.glob(r'/var/packages/MailPlus-Server/target/usr/bin/gsuite/*.egg'):
    sys.path.append(egg)

from googleapiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials
import googleapiclient.errors as errors

class GsuiteApi():
    def __init__(self, json_keyfile, target_user):
        SCOPES = ['https://mail.google.com']
        self._credentials = ServiceAccountCredentials.from_json_keyfile_name(json_keyfile, SCOPES)
        self.target_user = target_user
        #https://oauth2client.readthedocs.io/en/latest/source/oauth2client.service_account.html
        self._delegated_credentials = self._credentials.create_delegated(target_user)
        self._service = build('gmail', 'v1', credentials=self._delegated_credentials)

    def listlabels(self):
        # https://developers.google.com/gmail/api/v1/reference/users/labels/list
        return self._service.users().labels().list(userId='me').execute()

    def getlabels(self, labelId):
        # https://developers.google.com/gmail/api/v1/reference/users/labels/list
        return self._service.users().labels().get(userId='me', id=labelId).execute()

    def listmessages(self, labelId=None, query=None, pageToken=None):
        # https://developers.google.com/gmail/api/v1/reference/users/messages/list
        labelIds = [labelId] if labelId else None
        return self._service.users().messages().list(userId='me', labelIds=labelIds, q=query, pageToken=pageToken, maxResults=1000).execute()

    def getmessage(self, messageId):
        # https://developers.google.com/gmail/api/v1/reference/users/messages/get
        return self._service.users().messages().get(userId='me', id=messageId, format='raw').execute()

    def getservice(self):
        return self._service

SYNO_SPECIAL_FOLDERS = {
    'INBOX': {
        'result_folder_name': 'INBOX',
        'query': '-in:draft'
    },
    'DRAFT': {
        'result_folder_name': 'Drafts',
        'query': None
    },
    'SPAM': {
        'result_folder_name': 'Junk',
        'query': None
    },
    'TRASH': {
        'result_folder_name': 'Trash',
        'query': None
    },
    'SENT': {
        'result_folder_name': 'Sent',
        'query': None
    }
}

class GsuiteRepository(BaseRepository):
    def __init__(self, reposname, account):
        """Initialize a MaildirRepository object.  Takes a path name
        to the directory holding all the Maildir directories."""

        BaseRepository.__init__(self, reposname, account)

        self.folders = None
        self.ui = getglobalui()
        json_keyfile = self.getconf('json_keyfile', None)
        self.skip_archive = self.getconfboolean('skip_archived_mail', False)
        try:
            self._api = GsuiteApi(json_keyfile, self.getuser())
        except:
            self.ui.gsuite_json_keyfile_error()
            raise

        self.should_do_syno_nametrans = self.getconfboolean('should_do_syno_nametrans', False)
        if self.should_do_syno_nametrans:
            self._special_folders = dict()
            self._all_visible_folder_names = set()

        self.should_skip_trash = self.getconfboolean('skiptrash', False)
        self.should_skip_junk = self.getconfboolean('skipjunk', False)

    def getsep(self):
        return '/'

    def getuser(self):
        user = self.getconf('remoteuser', None)
        return user

    def getfolder(self, foldername):
        """Return a Folder instance of this Gsuite service

        If necessary, scan and cache all foldernames to make sure that
        we only return existing folders and that 2 calls with the same
        name will return the same object."""
        folders = self.getfolders()
        for f in folders:
            if f.name == foldername:
                return f

        raise OfflineImapError("getfolder() asked for a nonexisting "
            "folder '%s'."% foldername, OfflineImapError.ERROR.FOLDER)

    def _memorize_special_folder_name(self, folder):
        """memorize the name of the special folder to help us do syno nametrans"""
        #add by Synology, translate the folder name considering folder flags, dot, and inbox
        folder_id = folder.get_uidvalidity()
        if folder_id in SYNO_SPECIAL_FOLDERS:
            self._special_folders[folder_id] = {
                'name': folder.ffilter_name,
                'length': len(folder.ffilter_name)
            }

    def _do_syno_nametrans(self, folders):
        """How to translate the folder name by folder flags"""
        #add by Synology, translate the folder name considering folder id, dot, and inbox
        folder_separator = self.getsep()

        # If the result folder name also exists in the folder list, we do not do the conversion
        for special_folder_id, folder_name_item in self._special_folders.items():
            if SYNO_SPECIAL_FOLDERS[special_folder_id]['result_folder_name'] in self._all_visible_folder_names:
                self._special_folders.pop(special_folder_id)

        all_visible_folder_names = set()
        for folder in folders:
            # convert special folder to the result folder name
            for special_folder_id, folder_name_item in self._special_folders.items():
                if folder.ffilter_name == folder_name_item['name']:
                    folder.visiblename = SYNO_SPECIAL_FOLDERS[special_folder_id]['result_folder_name']
                    break
                elif folder.ffilter_name.startswith(folder_name_item['name'] + folder_separator):
                    folder.visiblename = SYNO_SPECIAL_FOLDERS[special_folder_id]['result_folder_name'] + folder.ffilter_name[folder_name_item['length']:]
                    break
            # convert the inbox and its subfolder to uppercase
            if 'inbox' == folder.ffilter_name.lower():
                folder.visiblename = 'INBOX'
            if folder.ffilter_name.lower().startswith('inbox' + folder_separator):
                folder.visiblename = 'INBOX' + folder.ffilter_name[5:]
            folder.visiblename = folder.visiblename.replace('.', '_')
            all_visible_folder_names.add(folder.visiblename)
        # handle the consecutive /
        for folder in folders:
            visiblename = folder.visiblename
            newname = visiblename
            idx = visiblename.rfind('/', 0, len(visiblename))
            while idx != -1:
                subname = visiblename[:idx]
                if subname not in all_visible_folder_names:
                    newname = subname + '_' + newname[idx+1:]
                elif idx + 1 < len(visiblename) and newname[idx+1] == '/':
                    newname = subname + '_' + newname[idx+1:]
                idx = visiblename.rfind('/', 0, idx)
            if newname.endswith('/'):
                newname = newname[:-1] + '_'
            folder.visiblename = newname

        # handle two folder is nametransed to the same name
        all_visiblename = set()
        for folder in folders:
            if folder.visiblename not in all_visiblename:
                all_visiblename.add(folder.visiblename)
            else:
                ori_visiblename = folder.visiblename
                new_name = ''
                i = 0
                while i < 1000000:
                    new_name = '{0}_{1}'.format(folder.visiblename, i)
                    if new_name not in all_visiblename:
                        break
                    i += 1
                folder.visiblename = new_name
                all_visiblename.add(folder.visiblename)
                for another_folder in folders:
                    #use ffilter_name to check hierarchy in original folder name
                    #use visiblename to check hierarchy in new folder name
                    if another_folder.ffilter_name.startswith(folder.ffilter_name + folder_separator) and \
                        another_folder.visiblename.startswith(ori_visiblename + folder_separator):
                        ori_another_visiblename = another_folder.visiblename
                        another_folder.visiblename = folder.visiblename + folder_separator + another_folder.visiblename[(len(ori_visiblename) + 1):]
                        if ori_another_visiblename in all_visiblename:
                            all_visiblename.remove(ori_another_visiblename)
                            all_visiblename.add(another_folder.visiblename)
        self._all_visible_folder_names = all_visiblename

    def getfolders(self):
        if self.folders is not None:
            return self.folders

        skipsyslabel = set(['CHAT', 'IMPORTANT', 'STARRED', 'UNREAD'])
        if self.should_skip_trash:
            skipsyslabel.add('TRASH')
        if self.should_skip_junk:
            skipsyslabel.add('SPAM')

        labels = self._api.listlabels().get('labels', [])

        folders = []
        for label in labels:
            if label['id'] == 'DRAFT':
                folder = self.getfoldertype()(self._api, label, self)
            else:
                #only migrate mail to draft and not other label
                folder = self.getfoldertype()(self._api, label, self, extra_query='-in:draft')
            folder.sync_this = folder.decide_should_sync_folder()
            if label['type'] == 'system':
                if label['id'].startswith('CATEGORY_') or label['id'] in skipsyslabel:
                    folder.sync_this = False
            folders.append(folder)
            if True == self.should_do_syno_nametrans:
                self._memorize_special_folder_name(folder)
                self._all_visible_folder_names.add(folder.visiblename)

        if self.should_do_syno_nametrans:
            self._do_syno_nametrans(folders)

        if not self.skip_archive:
            no_label = {
                'type': 'system',
                'labelListVisibility': 'labelHide',
                'messageListVisibility': 'hide',
                'id': '',
                'name': 'Archived'
            }
            #spam and trash is defaultly exluded from message list
            folder = self.getfoldertype()(self._api, no_label, self, extra_query='-has:userlabels -in:inbox -in:draft -in:sent -in:chat')
            folders.append(folder)

        self.folders = folders
        return self.folders

    def getfoldertype(self):
        return folder.Gsuite.GsuiteFolder

    def forgetfolders(self):
        """Forgets the cached list of folders, if any.  Useful to run
        after a sync run."""

        self.folders = None

    def testremoteserver(self, istestconnectivity):
        if istestconnectivity:
            return
        self._api.listlabels()

class GsuiteLabelRepository(GsuiteRepository):

    class GsuiteLabel(object):
        @staticmethod
        def sort_by_length(labels):
            return sorted(labels, key=lambda label: label.name)

        @staticmethod
        def build_labels_relationship(labels, are_labels_sorted=False):
            if are_labels_sorted:
                sorted_labels = labels
            else:
                sorted_labels = sort_by_length(labels)
            name_to_label_dict = dict()
            root_labels = list()

            for label in sorted_labels:
                has_parent = False
                parent_to_be = label.name
                ridx = parent_to_be.rfind('/')
                while ridx >= 0:
                    parent_to_be = parent_to_be[:ridx+1]
                    if parent_to_be in name_to_label_dict:
                        label.parent = name_to_label_dict[parent_to_be]
                        label.parent.children.add(label)
                        has_parent = True
                        break
                    parent_to_be = parent_to_be[:ridx]
                    if parent_to_be in name_to_label_dict:
                        label.parent = name_to_label_dict[parent_to_be]
                        label.parent.children.add(label)
                        has_parent = True
                        break
                    ridx = parent_to_be.rfind('/')

                name_to_label_dict[label.name] = label
                if has_parent:
                    label_short_name = ""
                    parent_name_len = len(parent_to_be)
                    if parent_to_be.endswith('/'):
                        label_short_name = label.visiblename[parent_name_len:]
                        label.sep_omitted = True
                    elif parent_name_len + 1 == len(label.name):
                        #child with only 1 / character under parent
                        label_short_name = label.name[-1]
                        label.sep_omitted = True
                    else:
                        label_short_name = label.name[parent_name_len + 1:]
                    label.short_name = label_short_name
                else:
                    root_labels.append(label)

            return root_labels

        def __init__(self, label_id, name, sync_this, color=None):
            self.mapped_keyword = ''
            self.parent = None
            self.children = set()
            self.id = label_id
            self.name = name
            self.visiblename = name
            self.short_name = name
            self.sync_this = sync_this
            self.sep_omitted = False
            self.color = color

    def __init__(self, reposname, account):
        """Initialize a MaildirRepository object.  Takes a path name
        to the directory holding all the Maildir directories."""
        import re

        GsuiteRepository.__init__(self, reposname, account)

        self._id_keyword_dict = dict()
        self.labels = None
        self.mailplus_user = self.getconf('mailplus_user', None)
        if self.mailplus_user is None:
            raise OfflineImapError('unknown mailplus user',
                OfflineImapError.ERROR.REPO)

        self.folder_filter_type = self.getconf('folderfiltertype', None)
        if self.folder_filter_type is not None:
            self.folder_filter_keyword = self.localeval.eval(
                self.getconf('folderinkeywordlist'), {'re': re})
            self.folder_filter_exception = self.localeval.eval(
                self.getconf('folderinexceptionlist'), {'re': re})
        else:
            self.folder_filter_keyword = lambda name : False
            self.folder_filter_exception = lambda name : False

    def _convert_label_for_mailplus(self):
        def convert_color_string(color_obj):
            #convert to hex value without # and all in uppercase
            if color_obj is None:
                #default color
                return 'DCE1E6', '50555A'
            else:
                return color_obj['backgroundColor'][1:].upper(), color_obj['textColor'][1:].upper()

        def recursive_add_label(json_labels, label_dict, label):

            item = dict()
            item['name'] = label.short_name
            if label.parent is not None:
                item['parent'] = label.parent.visiblename
                if not label.parent.sync_this:
                    recursive_add_label(json_labels, label_dict, label.parent)
            else:
                item['parent'] = ''

            item_key = (item['name'], item['parent'])
            if item_key in label_dict:
                return

            bg, text = convert_color_string(label.color)
            item['bg_color'] = bg
            item['text_color'] = text

            label_dict[item_key] = label
            json_labels.append(item)

        json_labels = list()
        label_dict = dict()

        for label in self.labels:
            if not label.sync_this:
                continue
            recursive_add_label(json_labels, label_dict, label)

        return label_dict, json_labels

    def sync_label_structure_to_mailplus(self):
        from subprocess import Popen, PIPE
        import json
        import shlex
        import subprocess

        #clear old mapping
        self._id_keyword_dict = dict()

        label_dict, json_labels = self._convert_label_for_mailplus()
        cmd = '/var/packages/MailClient/target/tool/label_migration -n "{0}"'.format(self.mailplus_user)
        try:
            cmd_list = shlex.split(cmd)
        except ValueError as e:
            raise OfflineImapError('Failed to sync label structure for account {0}'.format(self.account.getname()),
                OfflineImapError.ERROR.REPO)

        p = subprocess.Popen(cmd_list, shell=False,
            stdin=PIPE, stdout=PIPE, stderr=PIPE,
            close_fds=True)
        stdout, stderr = p.communicate(input=(json.dumps(json_labels, ensure_ascii=False).encode('utf8')))

        if p.returncode != 0:
            should_show_on_ui = False
            try:
                error_code = int(stdout)
                if error_code == 120:
                    err_msg = 'Invalid label from user [{0}]. Please contact Synology Technical Support.'.format(self.mailplus_user)
                elif error_code == 400:
                    err_msg = 'Unknown error from user [{0}]. Please contact Synology Technical Support.'.format(self.mailplus_user)
                elif error_code == 402:
                    should_show_on_ui = True
                    err_msg = 'User [{0}] is not activated or does not have permission to Synology MailPlus.'.format(self.mailplus_user)
                elif error_code == 407:
                    err_msg = 'The database of user [{0}] cannot be accessed. Please contact Synology Technical Support.'.format(self.mailplus_user)
                elif error_code == 410:
                    should_show_on_ui = True
                    err_msg = 'Upgrading the database of user [{0}]. Please try again later.'.format(self.mailplus_user)
                elif error_code == 414:
                    should_show_on_ui = True
                    err_msg = 'User [{0}] cannot be found. Please check your user list.'.format(self.mailplus_user)
                elif error_code == 440:
                    should_show_on_ui = True
                    err_msg = 'User [{0}] has a label longer than 256 characters. Please shorten it.'.format(self.getuser())
                elif error_code == 441:
                    err_msg = 'User [{0}] has duplicated labels. Please contact Synology Technical Support.'.format(self.mailplus_user)
                elif error_code == 442:
                    err_msg = 'Unable to create a label for user [{0}]. Please contact Synology Technical Support.'.format(self.mailplus_user)
                elif error_code == 443:
                    err_msg = 'Parent label from user [{0}] does not exist. Please contact Synology Technical Support.'.format(self.mailplus_user)
                else:
                    err_msg = 'Unknown error from user [{0}]. Please contact Synology Technical Support.'.format(self.mailplus_user)
                self.ui.sync_label_error('{0} ({1})'.format(err_msg, error_code), should_show_on_ui)
            except ValueError:
                self.ui.sync_label_error('Unknown error from user [{0}]. Please contact Synology Technical Support.'.format(self.mailplus_user))

            raise OfflineImapError("could not sync label structure to MailPlus", OfflineImapError.ERROR.REPO)

        try:
            mailplus_labels = json.loads(stdout)
        except:
            self.ui.sync_label_error('User {0} get wrong response when when sync label to MailPlus'.format(self.mailplus_user))
            raise

        for mailplus_label in mailplus_labels:
            if 'name' not in mailplus_label or \
               'parent' not in mailplus_label or \
               'keyword' not in mailplus_label:
                raise OfflineImapError("get wrong result while syncing label structure to MailPlus",
                    OfflineImapError.ERROR.REPO)
            key_tuple = (mailplus_label['name'], mailplus_label['parent'])
            if key_tuple not in label_dict:
                continue
            gmail_label = label_dict[key_tuple]
            gmail_label.mapped_keyword = mailplus_label['keyword']
            if gmail_label.sync_this:
                self._id_keyword_dict[gmail_label.id] = gmail_label.mapped_keyword

    def _handle_special_folder(self, folder):
        """convert the name of the special folder"""
        #add by Synology, translate the folder name considering folder flags, dot, and inbox
        folder_id = folder.get_uidvalidity()
        if folder_id in SYNO_SPECIAL_FOLDERS:
            folder.visiblename = SYNO_SPECIAL_FOLDERS[folder_id]['result_folder_name']
            query = SYNO_SPECIAL_FOLDERS[folder_id]['query']
            if query is not None:
                folder._extra_query = query

    def _do_syno_nametrans(self, root_labels):
        """How to translate the label name by label flags"""
        #add by Synology, translate the label name considering label id, dot, and inbox

        def _traverse_children_label(parent):
            #record known short name
            parent_visible_name_len = len(parent.visiblename)
            parent_name_changed = (parent.name != parent.visiblename)

            for label in parent.children:
                if parent_name_changed:
                    #parent name has been changed
                    label.visiblename = parent.visiblename + label.visiblename[len(parent.name):]
                if label.sep_omitted:
                    # separator is ommited, so we add it back
                    label.visiblename = parent.visiblename + '/' + label.visiblename[len(parent.visiblename):]

                if label.short_name.find('/') != -1:
                    result_name = label.short_name.replace('/', '|')
                    label.visiblename = label.visiblename[:-(len(label.short_name))] + result_name
                    label.short_name = result_name

                _traverse_children_label(label)

        dummy_root_label = self.GsuiteLabel(label_id='root_label', name='', sync_this=False, color=None)
        dummy_root_label.children = set(root_labels)
        _traverse_children_label(dummy_root_label)

    def getfolders(self):
        if self.folders is not None:
            return self.folders

        skip_labels = set(['CHAT', 'IMPORTANT', 'CATEGORY_FORUMS', 'CATEGORY_UPDATES', 'CATEGORY_PERSONAL', 'CATEGORY_PROMOTIONS', 'CATEGORY_SOCIAL', 'STARRED', 'UNREAD'])

        if self.should_skip_trash:
            skip_labels.add('TRASH')
        if self.should_skip_junk:
            skip_labels.add('SPAM')

        gsuite_labels = self._api.listlabels().get('labels', [])

        folders = list()
        labels = list()
        match_keyword_label_list = list()
        match_exception_label_list = list()
        filter_query = ''
        keyword_format = ''
        exception_format = ''
        query_format = ''
        add_extra_query = False
        if self.folder_filter_type == 'blacklist':
            keyword_format = u'{0} -label:{1}'
            exception_format = u'{0} label:{1}'
            query_format = u'{{({0}) {1}}}'
            exception_update_sync_this_fun = lambda sync_this : sync_this and True
            keyword_update_sync_this_fun = lambda sync_this : False
            add_extra_query = True
        elif self.folder_filter_type == 'whitelist':
            keyword_format = u'{0} label:{1}'
            exception_format = u'{0} -label:{1}'
            query_format = u'({{{0}}} {1})'
            exception_update_sync_this_fun = lambda sync_this : False
            keyword_update_sync_this_fun = lambda sync_this : sync_this and True
            add_extra_query = True
        else:
            exception_update_sync_this_fun = lambda sync_this : sync_this

        for label in gsuite_labels:
            if label['id'] in skip_labels:
                continue

            folder = self.getfoldertype()(self._api, label, self)
            if label['type'] != 'system':
                folder.sync_this = True
            if self.folder_filter_exception(label['name']):
                if label['type'] != 'system':
                    match_exception_label_list.append(label)
                folder.sync_this = exception_update_sync_this_fun(folder.sync_this)
            elif self.folder_filter_keyword(label['name']):
                if label['type'] != 'system':
                    match_keyword_label_list.append(label)
                folder.sync_this = keyword_update_sync_this_fun(folder.sync_this)
            elif label['type'] != 'system':
                folder.sync_this = exception_update_sync_this_fun(folder.sync_this)


            if label['type'] == 'system':
                folders.append(folder)
                self._handle_special_folder(folder)
            else:
                #user defined label
                gsuite_label = self.GsuiteLabel(label['id'], label['name'],\
                                                folder.sync_this,\
                                                label['color'] if 'color' in label else None)
                labels.append(gsuite_label)

        labels = self.GsuiteLabel.sort_by_length(labels)
        root_labels = self.GsuiteLabel.build_labels_relationship(labels, are_labels_sorted=True)
        if self.should_do_syno_nametrans:
            self._do_syno_nametrans(root_labels)

        if not self.skip_archive:
            archive_label = {
                'type': 'system',
                'labelListVisibility': 'labelHide',
                'messageListVisibility': 'hide',
                'id': '',
                'name': 'Archived'
            }
            #spam and trash is defaultly exluded from message list
            folder = self.getfoldertype()(self._api, archive_label, self, extra_query='-in:inbox -in:draft -in:sent -in:chat')
            folders.append(folder)

        #Only if some labels are matched we add extra query
        if match_keyword_label_list:
            import re
            keyword_query = reduce(lambda query_str, label : keyword_format.format(query_str, re.sub('[ &()-/{|}"]', '-', label['name'])), match_keyword_label_list, '')
            if match_exception_label_list:
                exception_query = reduce(lambda query_str, label : exception_format.format(query_str, re.sub('[ &()-/{|}"]', '-', label['name'])), match_exception_label_list, '')
            else:
                exception_query = ''
            filter_query = query_format.format(keyword_query, exception_query)

            for folder in folders:
                if folder.extra_query is None:
                    folder.extra_query = filter_query
                else:
                    folder.extra_query = u'{0} {1}'.format(folder.extra_query, filter_query)
        elif self.folder_filter_type == 'whitelist':
            #if it is white list and no label is matched, skip all folders
            for folder in folders:
                folder.sync_this = False

        self.labels = labels
        self.folders = folders
        return self.folders

    def forgetfolders(self):
        """Forgets the cached list of folders, if any.  Useful to run
        after a sync run."""

        self.folders = None
        self.labels = None
        self._id_keyword_dict = dict()

    def getfoldertype(self):
        return folder.Gsuite.GsuiteLabelFolder

    def convert_label_ids_to_keywords(self, label_ids):
        keywords_list = list()
        for label_id in label_ids:
            if label_id in self._id_keyword_dict:
                keywords_list.append(self._id_keyword_dict[label_id])
        return keywords_list

    def getfoldertype(self):
        return folder.Gsuite.GsuiteLabelFolder
